"""
Services defined by the Workspaces system, to be used by our APIs and client programs.
"""
import shutil
import stat
import subprocess
from pathlib import Path
from queue import PriorityQueue
from tempfile import mkdtemp
from typing import Dict, List, Union, Optional, Type

from sqlalchemy.orm import Session
from workflow.event_catcher import EventCatcher

from .capability_interfaces import CapabilityServiceIF, CapabilityQueueIF, CapabilityInfoIF, \
    CapabilityEngineIF, CapabilityName, ParameterIF
from .helpers import CapabilitySequence, ExecutionPriority, RequestState, ExecutionState
from .product_interfaces import FutureProductIF
from .workflow_interfaces import WorkflowServiceIF, WorkflowInfoIF
from .schema import Workflow, WorkflowEvent, WorkflowEventType, CapabilityRequest, Capability, \
    CapabilityExecution, get_engine, get_session_factory, WorkflowRequest, AbstractFile
from channels.amqp_helpers import workflow_events,capability_events, CAPABILITY_STATUS_EXCH
from wf_monitor.monitor import WorkflowMonitor, WORKFLOW_STATUS_EXCH, log_decorator_factory


class CapabilityService(CapabilityServiceIF):
    """
    The capability service: clients access this to request capability runs
    """
    def __init__(self, info: CapabilityInfoIF):
        self.execution_pool = []
        self.queues = {}
        self.capability_info = info

    def create_request(self,
                       capability_name: str,
                       parameters: List[ParameterIF]=None,
                       products: List[FutureProductIF]=None) -> "CapabilityRequestIF":
        self.capability_info.create_capability_request(capability_name, parameters, products)

    def run_capability(self, request: CapabilityRequest) -> CapabilityExecution:
        """
        Run a capability with the requested settings
        :param request: Capability request
        """
        # FIXME: need request versions in between requests and executions
        execution_record = self.capability_info.create_execution_record(request.id)
        self.execution_pool.append(execution_record)
        return execution_record

    def enqueue_execution(
            self,
            execution_record: CapabilityExecution,
            priority: int = ExecutionPriority.Default.value
    ):
        """
        Move execution record that is ready to execute a workflow into the appropriate capability
        queue
        """
        request = self.capability_info.lookup_entity(
            CapabilityRequest, execution_record.capability_request
        )
        capability = self.capability_info.lookup_entity(Capability, request.capability)

        # Get correct queue or initialize one
        queue = self.queues.get(capability.id, CapabilityQueue(capability.max_jobs))
        queue.enqueue(execution_record, priority)
        self.queues[capability.id] = queue

        # Remove execution record from pool
        self.execution_pool.remove(execution_record)


class CapabilityEngine(CapabilityEngineIF):
    """
    Executes a prepare and run workflow step of a capability
    """
    def __init__(self, execution: CapabilityExecution):
        self.execution = execution

    def execute(self):
        """
        Communicate with workflow service and send a request to run a workflow with given settings
        FIXME: Implement
        """
        # workflow_service.execute(workflow_name, arguments, files)


class CapabilityInfo(CapabilityInfoIF):
    """
    Interface for data access object that can look up and record information about capabilities
    and capability requests. Accesses and modifies the database.
    """
    def __init__(self, session: Session):
        self.session = session

    def create_capability(
            self,
            name: CapabilityName,
            steps: CapabilitySequence,
            max_jobs: int
    ) -> int:
        """
        Create new capability and save it in the database
        :param name: Name of new capability
        :param steps: List of capability steps required to execute the capability
        :param max_jobs: Max allowed number of concurrent executions of this kind of capability
        :return: Integer identifier for the capability
        """
        capability = Capability(name=name, steps=str(steps), max_jobs=max_jobs)
        return self.save_entity(capability)

    def create_capability_request(
            self,
            capability: "CapabilityIF",
            parameters: List[ParameterIF],
            future_products: List[FutureProductIF],
            versions: List[str]
    ) -> int:
        """
        Create new capability request and save it in the database
        :param capability_id: ID of the requested capability
        :param parameters: List
        :param future_products:
        :param versions:
        :return: Integer identifier of the request
        """
        # FIXME: Parameters, future products and versions need valid str reprs
        request = CapabilityRequest(
            state=RequestState.Ready.name,
            capability=capability.id,
            parameters=str(parameters),
            future_products=str(future_products),
            versions=str(versions)
        )
        return self.save_entity(request)

    def create_execution_record(self, request_id: int) -> int:
        """
        Create new execution record for a request and save it in the database
        :param request_id: ID of the capability request
        :return: Integer identifier for the record
        """
        record = CapabilityExecution(
            state=ExecutionState.Ready.name, capability_request=request_id, current_step=0
        )
        return self.save_entity(record)

    def lookup_entity(
            self,
            entity_id: int,
            entity_schema: Union[Type[Capability], Type[CapabilityRequest], Type[CapabilityExecution]]
    ) -> Optional[Union[Capability, CapabilityRequest, CapabilityExecution]]:
        """
        Look up entity in database and return object representation of it if found
        :param entity_id: ID of entity to be searched for
        :param entity_schema: Database schema of the entity
        :return: Object representation of entity if found, else None
        """
        return self.session.query(entity_schema).filter(entity_schema.id == entity_id)

    def save_entity(self, entity: Union[Capability, CapabilityRequest, CapabilityExecution]) -> int:
        """
        Save a given entity and return an integer identifier for it
        :param entity:   the entity to save
        :return:         the entity's identifier
        """
        self.session.add(entity)
        self.session.commit()

    def lookup_capability_request(self, request_id) -> "CapabilityRequestIF":
        return self.lookup_entity(request_id, CapabilityRequest)


class CapabilityQueue(CapabilityQueueIF):
    def __init__(self, max_concurrency: int):
        self.queue = PriorityQueue()
        self.max_concurrency = max_concurrency
        self.paused = False

    def enqueue(self, request: CapabilityRequest, priority: int):
        self.queue.put((priority, request))

    def dequeue(self) -> CapabilityRequest:
        return self.queue.get()

    def pause(self):
        self.paused = True

    def unpause(self):
        self.paused = False

    def __len__(self):
        return self.queue.qsize()


class WorkflowService(WorkflowServiceIF):
    def execute(self, workflow_name: str, argument: Dict, files: List[Path]):
        """
        Execute a workflow per the supplied parameters.
        """

        # 1. look up workflow, returns workflow
        info = WorkflowInfo(self).lookup_workflow_definition(workflow_name)

        # 2. create and save request, return request id
        record = WorkflowRequest().create_request(workflow_name, argument)

        # 3. render templates to files, returns list of rendered files
        contents = info.render_templates(argument, files)

        # 4. prepare files for condor execution
        temp_folder = self._prepare_files_for_condor(contents)

        # 5. execute condor and retrieve log file
        log_file = self._execute_prepared(temp_folder)

        # 6. start reading log file
        monitor = WorkflowMonitor(str(log_file))
        # parse events from log
        events = monitor.events
        with workflow_events:
            for e in events:
                # send amqp event and update database
                self.on_workflow_event(e, record, temp_folder)


    @staticmethod
    def _prepare_files_for_condor(files: List[AbstractFile]) -> Path:
        """
        Place the files for Condor into a new temp directory and returns the directory.

        :param files:  a dictionary of filename -> content
        :return:       a Path
        """
        # 1. create a temporary directory
        temp_folder = Path(mkdtemp(dir=str(Path.home() / "tmp")))

        # 2. spool each of the temp files to it
        for file in files:
            (temp_folder / file.filename).write_bytes(file.content)

        # 3. make any scripts in there executable
        for file in temp_folder.glob('*.sh'):
            file.chmod(file.stat().st_mode | stat.S_IEXEC)

        # finished, return folder
        return temp_folder

    @staticmethod
    def _execute_prepared(folder: Path) -> Path:
        """
        Execute HTCondor using the named folder as the source of the files.

        :param folder:  the path to the folder to execute
        :return:        the path to the log file
        """
        print(f'executing on folder {folder}')

        # some file in here should end in .dag; that file is our dagman input
        dagman = list(folder.glob('*.dag'))[0]

        # ensure the log file exists
        logfile = folder / 'condor.log'
        logfile.touch()

        # submit
        subprocess.run(['condor_submit_dag', str(dagman)], cwd=str(folder.absolute()))

        # return the logfile
        return logfile

    # The next few things that need to happen here are in response to
    # WorkflowEvents that we receive from wf_monitor as the workflow
    # execution evolves. So we have to set up listening at some point
    # in this class
    def __init__(self):
        # 1. Start listening for events from the wf_monitor stream
        # self.channel = channels.workflow_events.listen(self.on_workflow_event)
        raise NotImplementedError

    def on_workflow_event(self, event: WorkflowEvent, request_record: WorkflowRequest, tmp_folder: Path):
        # 1. log that we received this event, somehow
        # 2. update the WorkflowRequest record with the state we got
        # 3. do per-event-type stuff, such as level change events, database
        #    updates, and logging

        catcher = EventCatcher()

        decorated_workflow_send = log_decorator_factory('Sending Workflow Event...')(workflow_events.send)
        decorated_capability_send = log_decorator_factory('Sending Capability Event...')(capability_events.send)

        # 1. send amqp event to workflow channel
        decorated_workflow_send(event, WORKFLOW_STATUS_EXCH)

        # 2. update request record with new status
        print(f'Updating state on request {request_record.workflow_request_id} to {event.type.name}...')
        request_record.update_status(event.type.name)

        # 3. do per-event-type stuff
        if event.type == WorkflowEventType.SUBMITTED:
            # this would be a good place to emit a level-changed event
            # for the capability system
            # TODO: figure out shape of capability event from workflow event
            capability_change = catcher.update_capability_from_workflow(event)
            decorated_capability_send(capability_change, CAPABILITY_STATUS_EXCH)
            # TODO: Find out capability states and replace strings
            CapabilityRequest.update_status("WAITING")
        elif event.type == WorkflowEventType.EXECUTING:
            # another good place to emit a level-changed event
            capability_change = catcher.update_capability_from_workflow(event)
            decorated_capability_send(capability_change, CAPABILITY_STATUS_EXCH)
            CapabilityRequest.update_status("STARTED")
        elif event.type == WorkflowEventType.TERMINATED:
            # another level-change event opportunity
            capability_change = catcher.update_capability_from_workflow(event)
            decorated_capability_send(capability_change, CAPABILITY_STATUS_EXCH)
            CapabilityRequest.update_status("COMPLETE")

            # remove temp directory and rendered templates
            shutil.rmtree(tmp_folder, ignore_errors=False, onerror=None)
        else:
            # there is an event type of OTHER which we can basically
            # ignore for now
            capability_change = catcher.update_capability_from_workflow(event)
            decorated_capability_send(capability_change, CAPABILITY_STATUS_EXCH)
            CapabilityRequest.update_status("OTHER")

    def __exit__(self, exc_type, exc_val, exc_tb):
        # this method lets us use "with WorkflowService as ws" type syntax
        # we should close our AMQP channel gracefully here
        # self.channel.close()
        pass


class WorkflowInfo(WorkflowInfoIF):
    """
    Implements WorkflowInfo, backed by a relational database, using SQL Alchemy.
    """

    def __init__(self, session):
        self.session = session

    def lookup_workflow_definition(self, name: str) -> Workflow:
        return self.session.query(Workflow).get(name)

    def all_workflows(self) -> List[Workflow]:
        return self.session.query(Workflow).all()
